/* -----------------------------------------------------------------------------
Rah'Dicks Vectorscope (2007-03-12)
==================================

The code used herein is based on the original Sonique visual API example vis.
Use this vis for whatever you want - under one condition:
if it's for commercial use or public performance, I'd like to receive word
and/or pics from the action and be credited.
If you're making money with it, I'd be happy with a small donation. :-)

Thomas Radeke, 2007
http://www.rahdick.at/
----------------------------------------------------------------------------- */

#include <windows.h>
#include "vis.h"
#include <math.h>

// some global variables
int oldX, oldY, CenterX, CenterY, clippedX, clippedY, ampL, ampR;
unsigned long oldTime;
bool bgFade;
// Keep a copy of the waveform for the "sample & hold" functionality
// Unfortunately, the waveform data is only 8 bit.
unsigned char	oldWaveform[2][512];

const char* vis_name = "RD's Vectorscope";

// some colour constants
unsigned long clBlack = 0x00000000;
unsigned long clGreen = 0xff00ff00;
unsigned long clDarkGreen = 0xff004000;
unsigned long clRed = 0xffff0000;
unsigned long clDarkRed = 0xffbb0000;
unsigned long clGray = 0xff606060;
unsigned long clWhite = 0xffffffff;

// myQueryInterface not used since XMPlay does not support it
QueryInterface* myQueryInterface = NULL;

void Initialize(void);
BOOL Render(unsigned long* Video, int width, int height, int pitch, VisData* pVD);
BOOL SaveSettings(char* FileName);
BOOL OpenSettings(char* FileName);
BOOL Deinit();
BOOL Clicked(int x, int y, int buttons);
BOOL ReceiveQueryInterface(QueryInterface* Interface);

VisInfo plugin = {
	1,
	"RD's Vectorscope",
	VI_WAVEFORM|SONIQUEVISPROC,
	&Initialize,
	&Render,
	&SaveSettings,
	&OpenSettings,
	&Deinit,
	&Clicked,
	&ReceiveQueryInterface
};

DLLEXPORT VisInfo* QueryModule(void){
	return &plugin;
}

void Initialize(void){
	// initialize "Center" and "old"-coord vars
	CenterX = 0;
	CenterY = 0;
	oldX = CenterX;
	oldY = CenterY;
	oldTime = 0;
}

// Bresenham line drawing algorithm, taken from here:
// http://www.cs.unc.edu/~mcmillan/comp136/Lecture6/Lines.html
// TODO: implement a nice anti-aliased line algorithm, if it's not too slow
void line(int x0, int y0, int x1, int y1, unsigned long* Video, int pitch, unsigned long colour){
    int dy = y1 - y0;
    int dx = x1 - x0;
    int stepx, stepy;

    if (dy < 0) { dy = -dy;  stepy = -1; } else { stepy = 1; }
    if (dx < 0) { dx = -dx;  stepx = -1; } else { stepx = 1; }
    dy <<= 1;                                                  // dy is now 2*dy
    dx <<= 1;                                                  // dx is now 2*dx

    Video[pitch*y0+x0] = colour;
    if (dx > dy) {
        int fraction = dy - (dx >> 1);                         // same as 2*dy - dx
        while (x0 != x1) {
            if (fraction >= 0) {
                y0 += stepy;
                fraction -= dx;                                // same as fraction -= 2*dx
            }
            x0 += stepx;
            fraction += dy;                                    // same as fraction -= 2*dy
            Video[pitch*y0+x0] = colour;
        }
    } else {
        int fraction = dx - (dy >> 1);
        while (y0 != y1) {
            if (fraction >= 0) {
                x0 += stepx;
                fraction -= dy;
            }
            y0 += stepy;
            fraction += dx;
            Video[pitch*y0+x0] = colour;
        }
    }
}

BOOL Render(unsigned long* Video, int width, int height, int pitch, VisData* pVD){
	int i, c;
	
	// clear the screen, if bgFade is not set - see "initialize()"
	if(!bgFade) {
		for(i=0;i<(width*height);i++) {
			Video[i] = clBlack;
		}
	}
	
	// green background grid
	int linecount = 20;
	// horizontal lines
	for (i=1; i < linecount; i++) {
		line(0, height*i/linecount, width-1, height*i/linecount, Video, pitch, clDarkGreen);
	}
	// vertical lines
	for (i=1; i < linecount; i++) {
		line(width*i/linecount, 0, width*i/linecount, height-1, Video, pitch, clDarkGreen);
	}
	
	// horizontally and vertically centered gray lines
	line(width/2, 0, width/2, height-1, Video, pitch, clGray);
	line(0, height/2, width-1, height/2, Video, pitch, clGray);
	
	// diagonal red lines
	line(0, 0, width-1, height-1, Video, pitch, clDarkRed);
	line(0, height-1, width-1, 0, Video, pitch, clDarkRed);
	
	// get the screen center
	CenterX = width / 2;
	CenterY = height / 2;
	
	// draw the actual Vectorscope
	// iterate backwards through all of the waveform values for each frame
	// I do it backwards because I'm fading the brightness down from white
	// and don't want the darker lines to overwrite the brighter ones
	for(i = 511; i >= 0; i--) {
		
		// Sample and hold the data if stopped or paused
		if(oldTime < pVD->MillSec) {
			for(int j = 0; j < 512; j++) {
				oldWaveform[0][j] = pVD->Waveform[0][j];
				oldWaveform[1][j] = pVD->Waveform[1][j];
			}
		}
		// get left channel data
		ampL = oldWaveform[0][i];
		// the waveform data arrives as unsigned - all negative data is above 127
		if(ampL > 127) ampL = ampL-255;

		// the same for the right channel data		
		ampR = oldWaveform[1][i];
		if(ampR > 127) ampR = ampR-255;
		
		// calculate clipped and vertically centered vis data
		clippedY = (ampR*height)/255+CenterY;
		if(clippedY < 0) clippedY = 0;
		if(clippedY > (height-1)) clippedY = (height-1);
		
		// same again with the horizontally centered data
		clippedX = (-ampL*width)/255+CenterX;
		if(clippedX < 0) clippedX = 0;
		if(clippedX > width) clippedX = width-1;
		
		// avoid drawing of the first line
		if(i == 511) {
			oldX = clippedX;
			oldY = clippedY;
			//c = clRed;
		}
		else {
			// using the WinGDI RGB macro to convert the faded colour to unsigned long
			c = clGreen-RGB(0, i/4, 0);
		}
		
		// Now draw a line from the previous coordinates to the current ones
		line(oldX, oldY, clippedX, clippedY, Video, pitch, c);
		
		// Set "old" coords to current coords
		oldX = clippedX;
		oldY = clippedY;	
	}
	// get reference time
	oldTime = pVD->MillSec;

	return TRUE;
}

BOOL SaveSettings(char* FileName){
	char buffer[64];
	WritePrivateProfileString(vis_name, "FadeBackground", bgFade?"1":"0", FileName);
	return TRUE;
}

BOOL OpenSettings(char* FileName){
	char buffer[256];
	GetPrivateProfileString(vis_name, "FadeBackground", "0", buffer, sizeof(buffer), FileName);
	bgFade = (buffer[0]!='0');
	return TRUE;
}

BOOL Deinit()
{
	return TRUE;
}

BOOL Clicked(int x, int y, int buttons)
{
	bgFade^=true;
	return TRUE;
}

BOOL ReceiveQueryInterface(QueryInterface* Interface)
{
	myQueryInterface = Interface;
	return TRUE;
}
